package com.sun.xml.internal.ws.org.objectweb.asm;

import java.io.IOException;
import java.io.InputStream;

public class ClassReader
{
    static final boolean SIGNATURES = true;
    static final boolean ANNOTATIONS = true;
    static final boolean FRAMES = true;
    static final boolean WRITER = true;
    static final boolean RESIZE = true;
    public static final int SKIP_CODE = 1;
    public static final int SKIP_DEBUG = 2;
    public static final int SKIP_FRAMES = 4;
    public static final int EXPAND_FRAMES = 8;
    public final byte[] b;
    private final int[] items;
    private final String[] strings;
    private final int maxStringLength;
    public final int header;
    
    public ClassReader(byte[] b) {
        this(b, 0, b.length);
    }
    
    public ClassReader(byte[] b, int off, int len) {
        this.b = b;
        this.items = new int[this.readUnsignedShort(off + 8)];
        int n = this.items.length;
        this.strings = new String[n];
        int max = 0;
        int index = off + 10;
        int size;
        for (int i = 1; i < n; index += size, ++i) {
            this.items[i] = index + 1;
            switch (b[index]) {
                case 3:
                case 4:
                case 9:
                case 10:
                case 11:
                case 12: {
                    size = 5;
                    break;
                }
                case 5:
                case 6: {
                    size = 9;
                    ++i;
                    break;
                }
                case 1: {
                    size = 3 + this.readUnsignedShort(index + 1);
                    if (size > max) {
                        max = size;
                        break;
                    }
                    break;
                }
                default: {
                    size = 3;
                    break;
                }
            }
        }
        this.maxStringLength = max;
        this.header = index;
    }
    
    public int getAccess() {
        return this.readUnsignedShort(this.header);
    }
    
    public String getClassName() {
        return this.readClass(this.header + 2, new char[this.maxStringLength]);
    }
    
    public String getSuperName() {
        int n = this.items[this.readUnsignedShort(this.header + 4)];
        return (n == 0) ? null : this.readUTF8(n, new char[this.maxStringLength]);
    }
    
    public String[] getInterfaces() {
        int index = this.header + 6;
        int n = this.readUnsignedShort(index);
        String[] interfaces = new String[n];
        if (n > 0) {
            char[] buf = new char[this.maxStringLength];
            for (int i = 0; i < n; ++i) {
                index += 2;
                interfaces[i] = this.readClass(index, buf);
            }
            return interfaces;
        }
        return interfaces;
    }
    
    void copyPool(ClassWriter classWriter) {
        char[] buf = new char[this.maxStringLength];
        int ll = this.items.length;
        Item[] items2 = new Item[ll];
        Item item;
        int index2;
        for (int i = 1; i < ll; index2 = item.hashCode % items2.length, item.next = items2[index2], items2[index2] = item, ++i) {
            int index = this.items[i];
            int tag = this.b[index - 1];
            item = new Item(i);
            switch (tag) {
                case 9:
                case 10:
                case 11: {
                    int nameType = this.items[this.readUnsignedShort(index + 2)];
                    item.set(tag, this.readClass(index, buf), this.readUTF8(nameType, buf), this.readUTF8(nameType + 2, buf));
                    break;
                }
                case 3: {
                    item.set(this.readInt(index));
                    break;
                }
                case 4: {
                    item.set(Float.intBitsToFloat(this.readInt(index)));
                    break;
                }
                case 12: {
                    item.set(tag, this.readUTF8(index, buf), this.readUTF8(index + 2, buf), null);
                    break;
                }
                case 5: {
                    item.set(this.readLong(index));
                    ++i;
                    break;
                }
                case 6: {
                    item.set(Double.longBitsToDouble(this.readLong(index)));
                    ++i;
                    break;
                }
                case 1: {
                    String s = this.strings[i];
                    if (s == null) {
                        index = this.items[i];
                        String[] strings = this.strings;
                        int n = i;
                        String utf = this.readUTF(index + 2, this.readUnsignedShort(index), buf);
                        strings[n] = utf;
                        s = utf;
                    }
                    item.set(tag, s, null, null);
                    break;
                }
                default: {
                    item.set(tag, this.readUTF8(index, buf), null, null);
                    break;
                }
            }
        }
        int off = this.items[1] - 1;
        classWriter.pool.putByteArray(this.b, off, this.header - off);
        classWriter.items = items2;
        classWriter.threshold = (int)(0.75 * (double)ll);
        classWriter.index = ll;
    }
    
    public ClassReader(InputStream is) throws IOException {
        this(readClass(is));
    }
    
    public ClassReader(String name) throws IOException {
        this(ClassLoader.getSystemResourceAsStream(name.replace('.', '/') + ".class"));
    }
    
    private static byte[] readClass(InputStream is) throws IOException {
        if (is == null) {
            throw new IOException("Class not found");
        }
        byte[] b = new byte[is.available()];
        int len = 0;
        while (true) {
            int n = is.read(b, len, b.length - len);
            if (n == -1) {
                break;
            }
            len += n;
            if (len == b.length) {
                byte[] c = new byte[b.length + 1000];
                System.arraycopy(b, 0, c, 0, len);
                b = c;
            }
            else {
                continue;
            }
        }
        if (len < b.length) {
            byte[] c = new byte[len];
            System.arraycopy(b, 0, c, 0, len);
            b = c;
            return b;
        }
        return b;
    }
    
    public void accept(ClassVisitor classVisitor, int flags) {
        this.accept(classVisitor, new Attribute[0], flags);
    }
    
    public void accept(ClassVisitor classVisitor, Attribute[] attrs, int flags) {
        byte[] b = this.b;
        char[] c = new char[this.maxStringLength];
        int anns = 0;
        int ianns = 0;
        Attribute cattrs = null;
        int u = this.header;
        int access = this.readUnsignedShort(u);
        String name = this.readClass(u + 2, c);
        int v = this.items[this.readUnsignedShort(u + 4)];
        String superClassName = (v == 0) ? null : this.readUTF8(v, c);
        String[] implementedItfs = new String[this.readUnsignedShort(u + 6)];
        int w = 0;
        u += 8;
        for (int i = 0; i < implementedItfs.length; ++i) {
            implementedItfs[i] = this.readClass(u, c);
            u += 2;
        }
        boolean skipCode = (flags & 0x1) != 0x0;
        boolean skipDebug = (flags & 0x2) != 0x0;
        boolean unzip = (flags & 0x8) != 0x0;
        v = u;
        int i = this.readUnsignedShort(v);
        v += 2;
        while (i > 0) {
            int j = this.readUnsignedShort(v + 6);
            v += 8;
            while (j > 0) {
                v += 6 + this.readInt(v + 2);
                --j;
            }
            --i;
        }
        i = this.readUnsignedShort(v);
        v += 2;
        while (i > 0) {
            int j = this.readUnsignedShort(v + 6);
            v += 8;
            while (j > 0) {
                v += 6 + this.readInt(v + 2);
                --j;
            }
            --i;
        }
        String signature = null;
        String sourceFile = null;
        String sourceDebug = null;
        String enclosingOwner = null;
        String enclosingName = null;
        String enclosingDesc = null;
        for (i = this.readUnsignedShort(v), v += 2; i > 0; v += 6 + this.readInt(v + 2), --i) {
            String attrName = this.readUTF8(v, c);
            if ("SourceFile".equals(attrName)) {
                sourceFile = this.readUTF8(v + 6, c);
            }
            else if ("InnerClasses".equals(attrName)) {
                w = v + 6;
            }
            else if ("EnclosingMethod".equals(attrName)) {
                enclosingOwner = this.readClass(v + 6, c);
                int item = this.readUnsignedShort(v + 8);
                if (item != 0) {
                    enclosingName = this.readUTF8(this.items[item], c);
                    enclosingDesc = this.readUTF8(this.items[item] + 2, c);
                }
            }
            else if ("Signature".equals(attrName)) {
                signature = this.readUTF8(v + 6, c);
            }
            else if ("RuntimeVisibleAnnotations".equals(attrName)) {
                anns = v + 6;
            }
            else if ("Deprecated".equals(attrName)) {
                access |= 0x20000;
            }
            else if ("Synthetic".equals(attrName)) {
                access |= 0x1000;
            }
            else if ("SourceDebugExtension".equals(attrName)) {
                int len = this.readInt(v + 2);
                sourceDebug = this.readUTF(v + 6, len, new char[len]);
            }
            else if ("RuntimeInvisibleAnnotations".equals(attrName)) {
                ianns = v + 6;
            }
            else {
                Attribute attr = this.readAttribute(attrs, attrName, v + 6, this.readInt(v + 2), c, -1, (Label[])null);
                if (attr != null) {
                    attr.next = cattrs;
                    cattrs = attr;
                }
            }
        }
        classVisitor.visit(this.readInt(4), access, name, signature, superClassName, implementedItfs);
        if (!skipDebug && (sourceFile != null || sourceDebug != null)) {
            classVisitor.visitSource(sourceFile, sourceDebug);
        }
        if (enclosingOwner != null) {
            classVisitor.visitOuterClass(enclosingOwner, enclosingName, enclosingDesc);
        }
        for (i = 1; i >= 0; --i) {
            v = ((i == 0) ? ianns : anns);
            if (v != 0) {
                int j = this.readUnsignedShort(v);
                v += 2;
                while (j > 0) {
                    v = this.readAnnotationValues(v + 2, c, true, classVisitor.visitAnnotation(this.readUTF8(v, c), i != 0));
                    --j;
                }
            }
        }
        while (cattrs != null) {
            Attribute attr = cattrs.next;
            cattrs.next = null;
            classVisitor.visitAttribute(cattrs);
            cattrs = attr;
        }
        if (w != 0) {
            for (i = this.readUnsignedShort(w), w += 2; i > 0; w += 8, --i) {
                classVisitor.visitInnerClass((this.readUnsignedShort(w) == 0) ? null : this.readClass(w, c), (this.readUnsignedShort(w + 2) == 0) ? null : this.readClass(w + 2, c), (this.readUnsignedShort(w + 4) == 0) ? null : this.readUTF8(w + 4, c), this.readUnsignedShort(w + 6));
            }
        }
        i = this.readUnsignedShort(u);
        u += 2;
        while (i > 0) {
            access = this.readUnsignedShort(u);
            name = this.readUTF8(u + 2, c);
            String desc = this.readUTF8(u + 4, c);
            int fieldValueItem = 0;
            signature = null;
            anns = 0;
            ianns = 0;
            cattrs = null;
            for (int j = this.readUnsignedShort(u + 6), u += 8; j > 0; u += 6 + this.readInt(u + 2), --j) {
                String attrName = this.readUTF8(u, c);
                if ("ConstantValue".equals(attrName)) {
                    fieldValueItem = this.readUnsignedShort(u + 6);
                }
                else if ("Signature".equals(attrName)) {
                    signature = this.readUTF8(u + 6, c);
                }
                else if ("Deprecated".equals(attrName)) {
                    access |= 0x20000;
                }
                else if ("Synthetic".equals(attrName)) {
                    access |= 0x1000;
                }
                else if ("RuntimeVisibleAnnotations".equals(attrName)) {
                    anns = u + 6;
                }
                else if ("RuntimeInvisibleAnnotations".equals(attrName)) {
                    ianns = u + 6;
                }
                else {
                    Attribute attr = this.readAttribute(attrs, attrName, u + 6, this.readInt(u + 2), c, -1, (Label[])null);
                    if (attr != null) {
                        attr.next = cattrs;
                        cattrs = attr;
                    }
                }
            }
            FieldVisitor fv = classVisitor.visitField(access, name, desc, signature, (fieldValueItem == 0) ? null : this.readConst(fieldValueItem, c));
            if (fv != null) {
                for (int j = 1; j >= 0; --j) {
                    v = ((j == 0) ? ianns : anns);
                    if (v != 0) {
                        int k = this.readUnsignedShort(v);
                        v += 2;
                        while (k > 0) {
                            v = this.readAnnotationValues(v + 2, c, true, fv.visitAnnotation(this.readUTF8(v, c), j != 0));
                            --k;
                        }
                    }
                }
                while (cattrs != null) {
                    Attribute attr = cattrs.next;
                    cattrs.next = null;
                    fv.visitAttribute(cattrs);
                    cattrs = attr;
                }
                fv.visitEnd();
            }
            --i;
        }
        i = this.readUnsignedShort(u);
        u += 2;
        while (i > 0) {
            int u2 = u + 6;
            access = this.readUnsignedShort(u);
            name = this.readUTF8(u + 2, c);
            String desc = this.readUTF8(u + 4, c);
            signature = null;
            anns = 0;
            ianns = 0;
            int dann = 0;
            int mpanns = 0;
            int impanns = 0;
            cattrs = null;
            v = 0;
            w = 0;
            int attrSize;
            for (int j = this.readUnsignedShort(u + 6), u += 8; j > 0; u += attrSize, --j) {
                String attrName = this.readUTF8(u, c);
                attrSize = this.readInt(u + 2);
                u += 6;
                if ("Code".equals(attrName)) {
                    if (!skipCode) {
                        v = u;
                    }
                }
                else if ("Exceptions".equals(attrName)) {
                    w = u;
                }
                else if ("Signature".equals(attrName)) {
                    signature = this.readUTF8(u, c);
                }
                else if ("Deprecated".equals(attrName)) {
                    access |= 0x20000;
                }
                else if ("RuntimeVisibleAnnotations".equals(attrName)) {
                    anns = u;
                }
                else if ("AnnotationDefault".equals(attrName)) {
                    dann = u;
                }
                else if ("Synthetic".equals(attrName)) {
                    access |= 0x1000;
                }
                else if ("RuntimeInvisibleAnnotations".equals(attrName)) {
                    ianns = u;
                }
                else if ("RuntimeVisibleParameterAnnotations".equals(attrName)) {
                    mpanns = u;
                }
                else if ("RuntimeInvisibleParameterAnnotations".equals(attrName)) {
                    impanns = u;
                }
                else {
                    Attribute attr = this.readAttribute(attrs, attrName, u, attrSize, c, -1, (Label[])null);
                    if (attr != null) {
                        attr.next = cattrs;
                        cattrs = attr;
                    }
                }
            }
            String[] exceptions;
            if (w == 0) {
                exceptions = null;
            }
            else {
                exceptions = new String[this.readUnsignedShort(w)];
                w += 2;
                for (int j = 0; j < exceptions.length; ++j) {
                    exceptions[j] = this.readClass(w, c);
                    w += 2;
                }
            }
            MethodVisitor mv = classVisitor.visitMethod(access, name, desc, signature, exceptions);
            Label_5472: {
                if (mv != null) {
                    if ((mv instanceof MethodWriter)) {
                        MethodWriter mw = (MethodWriter)mv;
                        if (mw.cw.cr == this && signature == mw.signature) {
                            boolean sameExceptions = false;
                            if (exceptions == null) {
                                sameExceptions = (mw.exceptionCount == 0);
                            }
                            else if (exceptions.length == mw.exceptionCount) {
                                sameExceptions = true;
                                int j = exceptions.length - 1;
                                while (j >= 0) {
                                    w -= 2;
                                    if (mw.exceptions[j] != this.readUnsignedShort(w)) {
                                        sameExceptions = false;
                                        break;
                                    }
                                    --j;
                                }
                            }
                            if (sameExceptions) {
                                mw.classReaderOffset = u2;
                                mw.classReaderLength = u - u2;
                                break Label_5472;
                            }
                        }
                    }
                    if (dann != 0) {
                        AnnotationVisitor dv = mv.visitAnnotationDefault();
                        this.readAnnotationValue(dann, c, null, dv);
                        if (dv != null) {
                            dv.visitEnd();
                        }
                    }
                    for (int j = 1; j >= 0; --j) {
                        w = ((j == 0) ? ianns : anns);
                        if (w != 0) {
                            int k = this.readUnsignedShort(w);
                            w += 2;
                            while (k > 0) {
                                w = this.readAnnotationValues(w + 2, c, true, mv.visitAnnotation(this.readUTF8(w, c), j != 0));
                                --k;
                            }
                        }
                    }
                    if (mpanns != 0) {
                        this.readParameterAnnotations(mpanns, desc, c, true, mv);
                    }
                    if (impanns != 0) {
                        this.readParameterAnnotations(impanns, desc, c, false, mv);
                    }
                    while (cattrs != null) {
                        Attribute attr = cattrs.next;
                        cattrs.next = null;
                        mv.visitAttribute(cattrs);
                        cattrs = attr;
                    }
                }
                if (mv != null && v != 0) {
                    int maxStack = this.readUnsignedShort(v);
                    int maxLocals = this.readUnsignedShort(v + 2);
                    int codeLength = this.readInt(v + 4);
                    v += 8;
                    int codeStart = v;
                    int codeEnd = v + codeLength;
                    mv.visitCode();
                    Label[] labels = new Label[codeLength + 2];
                    this.readLabel(codeLength + 1, labels);
                    while (v < codeEnd) {
                        w = v - codeStart;
                        int opcode = b[v] & 0xFF;
                        switch (ClassWriter.TYPE[opcode]) {
                            case 0:
                            case 4: {
                                ++v;
                                continue;
                            }
                            case 8: {
                                this.readLabel(w + this.readShort(v + 1), labels);
                                v += 3;
                                continue;
                            }
                            case 9: {
                                this.readLabel(w + this.readInt(v + 1), labels);
                                v += 5;
                                continue;
                            }
                            case 16: {
                                opcode = (b[v + 1] & 0xFF);
                                if (opcode == 132) {
                                    v += 6;
                                    continue;
                                }
                                v += 4;
                                continue;
                            }
                            case 13: {
                                v = v + 4 - (w & 0x3);
                                this.readLabel(w + this.readInt(v), labels);
                                for (int j = this.readInt(v + 8) - this.readInt(v + 4) + 1, v += 12; j > 0; v += 4, --j) {
                                    this.readLabel(w + this.readInt(v), labels);
                                }
                                continue;
                            }
                            case 14: {
                                v = v + 4 - (w & 0x3);
                                this.readLabel(w + this.readInt(v), labels);
                                for (int j = this.readInt(v + 4), v += 8; j > 0; v += 8, --j) {
                                    this.readLabel(w + this.readInt(v + 4), labels);
                                }
                                continue;
                            }
                            case 1:
                            case 3:
                            case 10: {
                                v += 2;
                                continue;
                            }
                            case 2:
                            case 5:
                            case 6:
                            case 11:
                            case 12: {
                                v += 3;
                                continue;
                            }
                            case 7: {
                                v += 5;
                                continue;
                            }
                            default: {
                                v += 4;
                                continue;
                            }
                        }
                    }
                    for (int j = this.readUnsignedShort(v), v += 2; j > 0; v += 8, --j) {
                        Label start = this.readLabel(this.readUnsignedShort(v), labels);
                        Label end = this.readLabel(this.readUnsignedShort(v + 2), labels);
                        Label handler = this.readLabel(this.readUnsignedShort(v + 4), labels);
                        int type = this.readUnsignedShort(v + 6);
                        if (type == 0) {
                            mv.visitTryCatchBlock(start, end, handler, null);
                        }
                        else {
                            mv.visitTryCatchBlock(start, end, handler, this.readUTF8(this.items[type], c));
                        }
                    }
                    int varTable = 0;
                    int varTypeTable = 0;
                    int stackMap = 0;
                    int frameCount = 0;
                    int frameMode = 0;
                    int frameOffset = 0;
                    int frameLocalCount = 0;
                    int frameLocalDiff = 0;
                    int frameStackCount = 0;
                    Object[] frameLocal = null;
                    Object[] frameStack = null;
                    boolean zip = true;
                    cattrs = null;
                    for (int j = this.readUnsignedShort(v), v += 2; j > 0; v += 6 + this.readInt(v + 2), --j) {
                        String attrName = this.readUTF8(v, c);
                        if ("LocalVariableTable".equals(attrName)) {
                            if (!skipDebug) {
                                varTable = v + 6;
                                for (int k = this.readUnsignedShort(v + 6), w = v + 8; k > 0; w += 10, --k) {
                                    int label = this.readUnsignedShort(w);
                                    if (labels[label] == null) {
                                        Label label3 = this.readLabel(label, labels);
                                        label3.status |= 0x1;
                                    }
                                    label += this.readUnsignedShort(w + 2);
                                    if (labels[label] == null) {
                                        Label label4 = this.readLabel(label, labels);
                                        label4.status |= 0x1;
                                    }
                                }
                            }
                        }
                        else if ("LocalVariableTypeTable".equals(attrName)) {
                            varTypeTable = v + 6;
                        }
                        else if ("LineNumberTable".equals(attrName)) {
                            if (!skipDebug) {
                                int label;
                                for (int k = this.readUnsignedShort(v + 6), w = v + 8; k > 0; labels[label].line = this.readUnsignedShort(w + 2), w += 4, --k) {
                                    label = this.readUnsignedShort(w);
                                    if (labels[label] == null) {
                                        Label label2 = this.readLabel(label, labels);
                                        label2.status |= 0x1;
                                    }
                                }
                            }
                        }
                        else if ("StackMapTable".equals(attrName)) {
                            if ((flags & 0x4) == 0x0) {
                                stackMap = v + 8;
                                frameCount = this.readUnsignedShort(v + 6);
                            }
                        }
                        else if ("StackMap".equals(attrName)) {
                            if ((flags & 0x4) == 0x0) {
                                stackMap = v + 8;
                                frameCount = this.readUnsignedShort(v + 6);
                                zip = false;
                            }
                        }
                        else {
                            for (int k = 0; k < attrs.length; ++k) {
                                if (attrs[k].type.equals(attrName)) {
                                    Attribute attr = attrs[k].read(this, v + 6, this.readInt(v + 2), c, codeStart - 8, labels);
                                    if (attr != null) {
                                        attr.next = cattrs;
                                        cattrs = attr;
                                    }
                                }
                            }
                        }
                    }
                    if (stackMap != 0) {
                        frameLocal = new Object[maxLocals];
                        frameStack = new Object[maxStack];
                        Label_3700: {
                            if (unzip) {
                                int local = 0;
                                if ((access & 0x8) == 0x0) {
                                    if ("<init>".equals(name)) {
                                        frameLocal[local++] = Opcodes.UNINITIALIZED_THIS;
                                    }
                                    else {
                                        frameLocal[local++] = this.readClass(this.header + 2, c);
                                    }
                                }
                                int j = 1;
                                while (true) {
                                    int k = j;
                                    switch (desc.charAt(j++)) {
                                        case 'B':
                                        case 'C':
                                        case 'I':
                                        case 'S':
                                        case 'Z': {
                                            frameLocal[local++] = Opcodes.INTEGER;
                                            continue;
                                        }
                                        case 'F': {
                                            frameLocal[local++] = Opcodes.FLOAT;
                                            continue;
                                        }
                                        case 'J': {
                                            frameLocal[local++] = Opcodes.LONG;
                                            continue;
                                        }
                                        case 'D': {
                                            frameLocal[local++] = Opcodes.DOUBLE;
                                            continue;
                                        }
                                        case '[': {
                                            while (desc.charAt(j) == '[') {
                                                ++j;
                                            }
                                            if (desc.charAt(j) == 'L') {
                                                ++j;
                                                while (desc.charAt(j) != ';') {
                                                    ++j;
                                                }
                                            }
                                            frameLocal[local++] = desc.substring(k, ++j);
                                            continue;
                                        }
                                        case 'L': {
                                            while (desc.charAt(j) != ';') {
                                                ++j;
                                            }
                                            frameLocal[local++] = desc.substring(k + 1, j++);
                                            continue;
                                        }
                                        default: {
                                            frameLocalCount = local;
                                            break Label_3700;
                                        }
                                    }
                                }
                            }
                        }
                        frameOffset = -1;
                    }
                    v = codeStart;
                    while (v < codeEnd) {
                        w = v - codeStart;
                        Label l = labels[w];
                        if (l != null) {
                            mv.visitLabel(l);
                            if (!skipDebug && l.line > 0) {
                                mv.visitLineNumber(l.line, l);
                            }
                        }
                        while (frameLocal != null && (frameOffset == w || frameOffset == -1)) {
                            if (!zip || unzip) {
                                mv.visitFrame(-1, frameLocalCount, frameLocal, frameStackCount, frameStack);
                            }
                            else if (frameOffset != -1) {
                                mv.visitFrame(frameMode, frameLocalDiff, frameLocal, frameStackCount, frameStack);
                            }
                            if (frameCount > 0) {
                                int tag;
                                if (zip) {
                                    tag = (b[stackMap++] & 0xFF);
                                }
                                else {
                                    tag = 255;
                                    frameOffset = -1;
                                }
                                frameLocalDiff = 0;
                                int delta;
                                if (tag < 64) {
                                    delta = tag;
                                    frameMode = 3;
                                    frameStackCount = 0;
                                }
                                else if (tag < 128) {
                                    delta = tag - 64;
                                    stackMap = this.readFrameType(frameStack, 0, stackMap, c, labels);
                                    frameMode = 4;
                                    frameStackCount = 1;
                                }
                                else {
                                    delta = this.readUnsignedShort(stackMap);
                                    stackMap += 2;
                                    if (tag == 247) {
                                        stackMap = this.readFrameType(frameStack, 0, stackMap, c, labels);
                                        frameMode = 4;
                                        frameStackCount = 1;
                                    }
                                    else if (tag >= 248 && tag < 251) {
                                        frameMode = 2;
                                        frameLocalDiff = 251 - tag;
                                        frameLocalCount -= frameLocalDiff;
                                        frameStackCount = 0;
                                    }
                                    else if (tag == 251) {
                                        frameMode = 3;
                                        frameStackCount = 0;
                                    }
                                    else if (tag < 255) {
                                        int j = unzip ? frameLocalCount : 0;
                                        for (int k = tag - 251; k > 0; --k) {
                                            stackMap = this.readFrameType(frameLocal, j++, stackMap, c, labels);
                                        }
                                        frameMode = 1;
                                        frameLocalDiff = tag - 251;
                                        frameLocalCount += frameLocalDiff;
                                        frameStackCount = 0;
                                    }
                                    else {
                                        frameMode = 0;
                                        int n;
                                        frameLocalDiff = (n = (frameLocalCount = this.readUnsignedShort(stackMap)));
                                        stackMap += 2;
                                        int j = 0;
                                        while (n > 0) {
                                            stackMap = this.readFrameType(frameLocal, j++, stackMap, c, labels);
                                            --n;
                                        }
                                        frameStackCount = (n = this.readUnsignedShort(stackMap));
                                        stackMap += 2;
                                        j = 0;
                                        while (n > 0) {
                                            stackMap = this.readFrameType(frameStack, j++, stackMap, c, labels);
                                            --n;
                                        }
                                    }
                                }
                                frameOffset += delta + 1;
                                this.readLabel(frameOffset, labels);
                                --frameCount;
                            }
                            else {
                                frameLocal = null;
                            }
                        }
                        int opcode2 = b[v] & 0xFF;
                        switch (ClassWriter.TYPE[opcode2]) {
                            case 0: {
                                mv.visitInsn(opcode2);
                                ++v;
                                continue;
                            }
                            case 4: {
                                if (opcode2 > 54) {
                                    opcode2 -= 59;
                                    mv.visitVarInsn(54 + (opcode2 >> 2), opcode2 & 0x3);
                                }
                                else {
                                    opcode2 -= 26;
                                    mv.visitVarInsn(21 + (opcode2 >> 2), opcode2 & 0x3);
                                }
                                ++v;
                                continue;
                            }
                            case 8: {
                                mv.visitJumpInsn(opcode2, labels[w + this.readShort(v + 1)]);
                                v += 3;
                                continue;
                            }
                            case 9: {
                                mv.visitJumpInsn(opcode2 - 33, labels[w + this.readInt(v + 1)]);
                                v += 5;
                                continue;
                            }
                            case 16: {
                                opcode2 = (b[v + 1] & 0xFF);
                                if (opcode2 == 132) {
                                    mv.visitIincInsn(this.readUnsignedShort(v + 2), (int)this.readShort(v + 4));
                                    v += 6;
                                    continue;
                                }
                                mv.visitVarInsn(opcode2, this.readUnsignedShort(v + 2));
                                v += 4;
                                continue;
                            }
                            case 13: {
                                v = v + 4 - (w & 0x3);
                                int label = w + this.readInt(v);
                                int min = this.readInt(v + 4);
                                int max = this.readInt(v + 8);
                                v += 12;
                                Label[] table = new Label[max - min + 1];
                                for (int j = 0; j < table.length; ++j) {
                                    table[j] = labels[w + this.readInt(v)];
                                    v += 4;
                                }
                                mv.visitTableSwitchInsn(min, max, labels[label], table);
                                continue;
                            }
                            case 14: {
                                v = v + 4 - (w & 0x3);
                                int label = w + this.readInt(v);
                                int j = this.readInt(v + 4);
                                v += 8;
                                int[] keys = new int[j];
                                Label[] values = new Label[j];
                                for (j = 0; j < keys.length; ++j) {
                                    keys[j] = this.readInt(v);
                                    values[j] = labels[w + this.readInt(v + 4)];
                                    v += 8;
                                }
                                mv.visitLookupSwitchInsn(labels[label], keys, values);
                                continue;
                            }
                            case 3: {
                                mv.visitVarInsn(opcode2, b[v + 1] & 0xFF);
                                v += 2;
                                continue;
                            }
                            case 1: {
                                mv.visitIntInsn(opcode2, (int)b[v + 1]);
                                v += 2;
                                continue;
                            }
                            case 2: {
                                mv.visitIntInsn(opcode2, (int)this.readShort(v + 1));
                                v += 3;
                                continue;
                            }
                            case 10: {
                                mv.visitLdcInsn(this.readConst(b[v + 1] & 0xFF, c));
                                v += 2;
                                continue;
                            }
                            case 11: {
                                mv.visitLdcInsn(this.readConst(this.readUnsignedShort(v + 1), c));
                                v += 3;
                                continue;
                            }
                            case 6:
                            case 7: {
                                int cpIndex = this.items[this.readUnsignedShort(v + 1)];
                                String iowner = this.readClass(cpIndex, c);
                                cpIndex = this.items[this.readUnsignedShort(cpIndex + 2)];
                                String iname = this.readUTF8(cpIndex, c);
                                String idesc = this.readUTF8(cpIndex + 2, c);
                                if (opcode2 < 182) {
                                    mv.visitFieldInsn(opcode2, iowner, iname, idesc);
                                }
                                else {
                                    mv.visitMethodInsn(opcode2, iowner, iname, idesc);
                                }
                                if (opcode2 == 185) {
                                    v += 5;
                                    continue;
                                }
                                v += 3;
                                continue;
                            }
                            case 5: {
                                mv.visitTypeInsn(opcode2, this.readClass(v + 1, c));
                                v += 3;
                                continue;
                            }
                            case 12: {
                                mv.visitIincInsn(b[v + 1] & 0xFF, (int)b[v + 2]);
                                v += 3;
                                continue;
                            }
                            default: {
                                mv.visitMultiANewArrayInsn(this.readClass(v + 1, c), b[v + 3] & 0xFF);
                                v += 4;
                                continue;
                            }
                        }
                    }
                    Label l = labels[codeEnd - codeStart];
                    if (l != null) {
                        mv.visitLabel(l);
                    }
                    if (!skipDebug && varTable != 0) {
                        int[] typeTable = null;
                        if (varTypeTable != 0) {
                            int k;
                            for (k = this.readUnsignedShort(varTypeTable) * 3, w = varTypeTable + 2, typeTable = new int[k]; k > 0; typeTable[--k] = w + 6, typeTable[--k] = this.readUnsignedShort(w + 8), typeTable[--k] = this.readUnsignedShort(w), w += 10) {}
                        }
                        int k;
                        for (k = this.readUnsignedShort(varTable), w = varTable + 2; k > 0; w += 10, --k) {
                            int start2 = this.readUnsignedShort(w);
                            int length = this.readUnsignedShort(w + 2);
                            int index = this.readUnsignedShort(w + 8);
                            String vsignature = null;
                            if (typeTable != null) {
                                int a = 0;
                                while (a < typeTable.length) {
                                    if (typeTable[a] == start2 && typeTable[a + 1] == index) {
                                        vsignature = this.readUTF8(typeTable[a + 2], c);
                                        break;
                                    }
                                    a += 3;
                                }
                            }
                            mv.visitLocalVariable(this.readUTF8(w + 4, c), this.readUTF8(w + 6, c), vsignature, labels[start2], labels[start2 + length], index);
                        }
                    }
                    while (cattrs != null) {
                        Attribute attr = cattrs.next;
                        cattrs.next = null;
                        mv.visitAttribute(cattrs);
                        cattrs = attr;
                    }
                    mv.visitMaxs(maxStack, maxLocals);
                }
                if (mv != null) {
                    mv.visitEnd();
                }
            }
            --i;
        }
        classVisitor.visitEnd();
    }
    
    private void readParameterAnnotations(int v, String desc, char[] buf, boolean visible, MethodVisitor mv) {
        int n = this.b[v++] & 0xFF;
        int synthetics;
        int i;
        for (synthetics = Type.getArgumentTypes(desc).length - n, i = 0; i < synthetics; ++i) {
            AnnotationVisitor av = mv.visitParameterAnnotation(i, "Ljava/lang/Synthetic;", false);
            if (av != null) {
                av.visitEnd();
            }
        }
        while (i < n + synthetics) {
            int j = this.readUnsignedShort(v);
            v += 2;
            while (j > 0) {
                AnnotationVisitor av = mv.visitParameterAnnotation(i, this.readUTF8(v, buf), visible);
                v = this.readAnnotationValues(v + 2, buf, true, av);
                --j;
            }
            ++i;
        }
    }
    
    private int readAnnotationValues(int v, char[] buf, boolean named, AnnotationVisitor av) {
        int i = this.readUnsignedShort(v);
        v += 2;
        if (named) {
            while (i > 0) {
                v = this.readAnnotationValue(v + 2, buf, this.readUTF8(v, buf), av);
                --i;
            }
        }
        else {
            while (i > 0) {
                v = this.readAnnotationValue(v, buf, null, av);
                --i;
            }
        }
        if (av != null) {
            av.visitEnd();
        }
        return v;
    }
    
    private int readAnnotationValue(int v, char[] buf, String name, AnnotationVisitor av) {
        if (av == null) {
            switch (this.b[v] & 0xFF) {
                case 101: {
                    return v + 5;
                }
                case 64: {
                    return this.readAnnotationValues(v + 3, buf, true, null);
                }
                case 91: {
                    return this.readAnnotationValues(v + 1, buf, false, null);
                }
                default: {
                    return v + 3;
                }
            }
        }
        else {
            switch (this.b[v++] & 0xFF) {
                case 68:
                case 70:
                case 73:
                case 74: {
                    av.visit(name, this.readConst(this.readUnsignedShort(v), buf));
                    v += 2;
                    break;
                }
                case 66: {
                    av.visit(name, new Byte((byte)this.readInt(this.items[this.readUnsignedShort(v)])));
                    v += 2;
                    break;
                }
                case 90: {
                    av.visit(name, (this.readInt(this.items[this.readUnsignedShort(v)]) == 0) ? Boolean.FALSE : Boolean.TRUE);
                    v += 2;
                    break;
                }
                case 83: {
                    av.visit(name, new Short((short)this.readInt(this.items[this.readUnsignedShort(v)])));
                    v += 2;
                    break;
                }
                case 67: {
                    av.visit(name, new Character((char)this.readInt(this.items[this.readUnsignedShort(v)])));
                    v += 2;
                    break;
                }
                case 115: {
                    av.visit(name, this.readUTF8(v, buf));
                    v += 2;
                    break;
                }
                case 101: {
                    av.visitEnum(name, this.readUTF8(v, buf), this.readUTF8(v + 2, buf));
                    v += 4;
                    break;
                }
                case 99: {
                    av.visit(name, Type.getType(this.readUTF8(v, buf)));
                    v += 2;
                    break;
                }
                case 64: {
                    v = this.readAnnotationValues(v + 2, buf, true, av.visitAnnotation(name, this.readUTF8(v, buf)));
                    break;
                }
                case 91: {
                    int size = this.readUnsignedShort(v);
                    v += 2;
                    if (size == 0) {
                        return this.readAnnotationValues(v - 2, buf, false, av.visitArray(name));
                    }
                    switch (this.b[v++] & 0xFF) {
                        case 66: {
                            byte[] bv = new byte[size];
                            for (int i = 0; i < size; ++i) {
                                bv[i] = (byte)this.readInt(this.items[this.readUnsignedShort(v)]);
                                v += 3;
                            }
                            av.visit(name, bv);
                            --v;
                            break;
                        }
                        case 90: {
                            boolean[] zv = new boolean[size];
                            for (int i = 0; i < size; ++i) {
                                zv[i] = (this.readInt(this.items[this.readUnsignedShort(v)]) != 0);
                                v += 3;
                            }
                            av.visit(name, zv);
                            --v;
                            break;
                        }
                        case 83: {
                            short[] sv = new short[size];
                            for (int i = 0; i < size; ++i) {
                                sv[i] = (short)this.readInt(this.items[this.readUnsignedShort(v)]);
                                v += 3;
                            }
                            av.visit(name, sv);
                            --v;
                            break;
                        }
                        case 67: {
                            char[] cv = new char[size];
                            for (int i = 0; i < size; ++i) {
                                cv[i] = (char)this.readInt(this.items[this.readUnsignedShort(v)]);
                                v += 3;
                            }
                            av.visit(name, cv);
                            --v;
                            break;
                        }
                        case 73: {
                            int[] iv = new int[size];
                            for (int i = 0; i < size; ++i) {
                                iv[i] = this.readInt(this.items[this.readUnsignedShort(v)]);
                                v += 3;
                            }
                            av.visit(name, iv);
                            --v;
                            break;
                        }
                        case 74: {
                            long[] lv = new long[size];
                            for (int i = 0; i < size; ++i) {
                                lv[i] = this.readLong(this.items[this.readUnsignedShort(v)]);
                                v += 3;
                            }
                            av.visit(name, lv);
                            --v;
                            break;
                        }
                        case 70: {
                            float[] fv = new float[size];
                            for (int i = 0; i < size; ++i) {
                                fv[i] = Float.intBitsToFloat(this.readInt(this.items[this.readUnsignedShort(v)]));
                                v += 3;
                            }
                            av.visit(name, fv);
                            --v;
                            break;
                        }
                        case 68: {
                            double[] dv = new double[size];
                            for (int i = 0; i < size; ++i) {
                                dv[i] = Double.longBitsToDouble(this.readLong(this.items[this.readUnsignedShort(v)]));
                                v += 3;
                            }
                            av.visit(name, dv);
                            --v;
                            break;
                        }
                        default: {
                            v = this.readAnnotationValues(v - 3, buf, false, av.visitArray(name));
                            break;
                        }
                    }
                    break;
                }
            }
            return v;
        }
    }
    
    private int readFrameType(Object[] frame, int index, int v, char[] buf, Label[] labels) {
        int type = this.b[v++] & 0xFF;
        switch (type) {
            case 0: {
                frame[index] = Opcodes.TOP;
                break;
            }
            case 1: {
                frame[index] = Opcodes.INTEGER;
                break;
            }
            case 2: {
                frame[index] = Opcodes.FLOAT;
                break;
            }
            case 3: {
                frame[index] = Opcodes.DOUBLE;
                break;
            }
            case 4: {
                frame[index] = Opcodes.LONG;
                break;
            }
            case 5: {
                frame[index] = Opcodes.NULL;
                break;
            }
            case 6: {
                frame[index] = Opcodes.UNINITIALIZED_THIS;
                break;
            }
            case 7: {
                frame[index] = this.readClass(v, buf);
                v += 2;
                break;
            }
            default: {
                frame[index] = this.readLabel(this.readUnsignedShort(v), labels);
                v += 2;
                break;
            }
        }
        return v;
    }
    
    protected Label readLabel(int offset, Label[] labels) {
        if (labels[offset] == null) {
            labels[offset] = new Label();
        }
        return labels[offset];
    }
    
    private Attribute readAttribute(Attribute[] attrs, String type, int off, int len, char[] buf, int codeOff, Label[] labels) {
        int i = 0;
        while (i < attrs.length) {
            if (attrs[i].type.equals(type)) {
                return attrs[i].read(this, off, len, buf, codeOff, labels);
            }
            ++i;
        }
        return new Attribute(type).read(this, off, len, null, -1, (Label[])null);
    }
    
    public int getItem(int item) {
        return this.items[item];
    }
    
    public int readByte(int index) {
        return this.b[index] & 0xFF;
    }
    
    public int readUnsignedShort(int index) {
        byte[] b = this.b;
        return (b[index] & 0xFF) << 8 | (b[index + 1] & 0xFF);
    }
    
    public short readShort(int index) {
        byte[] b = this.b;
        return (short)((b[index] & 0xFF) << 8 | (b[index + 1] & 0xFF));
    }
    
    public int readInt(int index) {
        byte[] b = this.b;
        return (b[index] & 0xFF) << 24 | (b[index + 1] & 0xFF) << 16 | (b[index + 2] & 0xFF) << 8 | (b[index + 3] & 0xFF);
    }
    
    public long readLong(int index) {
        long l1 = (long)this.readInt(index);
        long l2 = (long)this.readInt(index + 4) & 0xFFFFFFFFL;
        return l1 << 32 | l2;
    }
    
    public String readUTF8(int index, char[] buf) {
        int item = this.readUnsignedShort(index);
        String s = this.strings[item];
        if (s != null) {
            return s;
        }
        index = this.items[item];
        return this.strings[item] = this.readUTF(index + 2, this.readUnsignedShort(index), buf);
    }
    
    private String readUTF(int index, int utfLen, char[] buf) {
        int endIndex = index + utfLen;
        byte[] b = this.b;
        int strLen = 0;
        while (index < endIndex) {
            int c = b[index++] & 0xFF;
            switch (c >> 4) {
                case 0:
                case 1:
                case 2:
                case 3:
                case 4:
                case 5:
                case 6:
                case 7: {
                    buf[strLen++] = (char)c;
                    continue;
                }
                case 12:
                case 13: {
                    int d = b[index++];
                    buf[strLen++] = (char)((c & 0x1F) << 6 | (d & 0x3F));
                    continue;
                }
                default: {
                    int d = b[index++];
                    int e = b[index++];
                    buf[strLen++] = (char)((c & 0xF) << 12 | (d & 0x3F) << 6 | (e & 0x3F));
                    continue;
                }
            }
        }
        return new String(buf, 0, strLen);
    }
    
    public String readClass(int index, char[] buf) {
        return this.readUTF8(this.items[this.readUnsignedShort(index)], buf);
    }
    
    public Object readConst(int item, char[] buf) {
        int index = this.items[item];
        switch (this.b[index - 1]) {
            case 3: {
                return new Integer(this.readInt(index));
            }
            case 4: {
                return new Float(Float.intBitsToFloat(this.readInt(index)));
            }
            case 5: {
                return new Long(this.readLong(index));
            }
            case 6: {
                return new Double(Double.longBitsToDouble(this.readLong(index)));
            }
            case 7: {
                return Type.getObjectType(this.readUTF8(index, buf));
            }
            default: {
                return this.readUTF8(index, buf);
            }
        }
    }
}
